# -*- coding: utf-8 -*-
import re
import requests
from bs4 import BeautifulSoup
from bs4.element import Tag
from django.utils.translation import get_language


class BaseDataProvider(object):
    """ Basic data provider
    with base methods, helpers and universal parsers
    """
    # base models
    line_model = None
    station_model = None
    # various parser stuff
    bg_word = 'background:'
    without_brackets_re = re.compile(ur'\s*\(.*\)')
    lang = None

    """ main methods
    """
    def __init__(self, station_model=None, line_model=None):
        if station_model and line_model:
            self.station_model = station_model
            self.line_model = line_model
            self.lang = get_language()
        else:
            raise AttributeError(
                u'You need to provide base station and line models'
            )

    def download_all(self):
        self.download_lines()
        self.download_stations()

    def download_lines(self):
        pass

    def download_stations(self):
        pass

    """ helper methods
    """
    def create_dom(self, url):
        get = lambda: BeautifulSoup(requests.get(url).content)
        try:
            html = get()
        except:
            # one more try
            try:
                html = get()
            except:
                return None
        return html

    def get_or_create_line(self, i, el_or_title, color=None):
        is_tag = isinstance(el_or_title, Tag)
        line, _ = self.line_model.objects.get_or_create(
            **dict(
                number=i,
                defaults=dict(
                    title=self.prep_title(el_or_title),
                    color=self.prep_color(el_or_title if is_tag else color),
                )
            )
        )
        return line

    def get_or_create_station(self, line, el):
        station, _ = self.station_model.objects.get_or_create(
            line=line,
            title=self.prep_title(self.get_el_text(el)),
        )
        return station

    def get_el_text(self, el):
        return el.get_text() if isinstance(el, Tag) else unicode(el)

    def prep_color(self, el):
        try:
            element = el['style'] if isinstance(el, Tag) else el
            return element\
                .replace(self.bg_word, '')\
                .replace(';', '')\
                .strip()
        except KeyError:
            return ''


class BaseRuDataProvider(BaseDataProvider):
    """ Basic data provider for ru lang
    """
    line_word = u'линия'
    clean_title_re = re.compile(ur'[^а-яА-ЯёЁa-zA-Z0-9\s\.,-_]*', re.U | re.I)

    def prep_title(self, el):
        return self.translit_if_needed(
            self.clean_title_re.sub(
                '',
                self.without_brackets_re.sub(
                    '',
                    self.get_el_text(el)\
                        .replace(self.line_word, '')\
                        .strip()
                )
            )
        )

    def translit_if_needed(self, el):
        if 'ru' not in self.lang:
            return self.transliterate(el)
        return el

    def transliterate(self, s):
        # reduce requirements
        repl_dict = {
            # upper
            u'А': u'A', u'Б': u'B', u'В': u'V', u'Г': u'G', u'Д': u'D',
            u'Е': u'E', u'Ё': u'Yo', u'З': u'Z', u'И': u'I', u'Й': u'Y',
            u'К': u'K', u'Л': u'L', u'М': u'M', u'Н': u'N', u'О': u'O',
            u'П': u'P', u'Р': u'R', u'С': u'S', u'Т': u'T', u'У': u'U',
            u'Ф': u'F', u'Х': u'H', u'Ъ': u'', u'Ы': u'Y', u'Ь': u'',
            u'Э': u'E', u'Ж': u'Zh',  u'Ц': u'Ts', u'Ч': u'Ch', u'Ш': u'Sh',
            u'Щ': u'Sch', u'Ю': u'Yu', u'Я': u'Ya',
            # lower
            u'а': u'a', u'б': u'b', u'в': u'v', u'г': u'g',
            u'д': u'd', u'е': u'e', u'ё': u'yo', u'ж': u'zh', u'з': u'z',
            u'и': u'i', u'й': u'y', u'к': u'k', u'л': u'l', u'м': u'm',
            u'н': u'n', u'о': u'o', u'п': u'p', u'р': u'r', u'с': u's',
            u'т': u't', u'у': u'u', u'ф': u'f', u'х': u'h', u'ц': u'ts',
            u'ч': u'ch', u'ш': u'sh', u'щ': u'sch', u'ъ': u'', u'ы': u'y',
            u'ь': u'', u'э': u'e', u'ю': u'yu', u'я': u'ya',
        }
        for key in repl_dict:
            s = s.replace(key, repl_dict[key])
        return s

    """ universal cases
    """
    # case with one big table, containing all lines and stations
    def parse_usual_big_table(self, station_td_count=None, table_number=None):
        i = 0
        line = None
        html = self.create_dom(self.metro_data_src)
        # search only in body
        body = html.find(id='mw-content-text')
        if table_number:
            table = body.find_all('table')[table_number]
        else:
            table = body.find('table')
        # get tbody instead of table, if exist
        tbody = table.find('tbody', recursive=False)
        if tbody:
            table = tbody
        stations = []
        for tr in table.find_all('tr', recursive=False):
            td = tr.find('td', recursive=False)
            if td:
                if 'rowspan' in td.attrs:
                    continue
                if self.line_word in td.get_text():
                    i += 1
                    line = self.get_or_create_line(i, td)
                elif line:
                    if 'colspan' in td.attrs:
                        break
                    # additional check for some cases
                    if station_td_count:
                        count = len(
                            tr.find_all('td', recursive=False)
                        )
                        if count != station_td_count:
                            continue
                    self.get_or_create_station(line, td)


class BaseEnDataProvider(BaseDataProvider):
    """ Basic data provider for en lang
    """
    base_en_url = u'http://en.wikipedia.org'
    without_sbrackets_re = re.compile(ur'\s*\[.*\]')

    def prep_title(self, el):
        return self.without_brackets_re.sub(
            '', self.without_sbrackets_re.sub(
                '', self.get_el_text(el).strip()
            )
        )



# -*- coding: utf-8 -*-
import re
import requests
from bs4 import BeautifulSoup

from metro.parser.base import BaseRuDataProvider


class DataProvider(BaseRuDataProvider):
    metro_data_src = u"http://ru.wikipedia.org/wiki/\
                       Список_станций_Екатеринбургского_метрополитена"
    unknown = u'неизвестн'

    def download_all(self):
        html = self.create_dom(self.metro_data_src)
        # i dont want to parse this station, im tired
        line = self.get_or_create_line(1, u'Первая', 'green')
        # get stations
        table = html.find('table', class_='wide')
        for cell in table.find_all('tr'):
            tds = cell.find_all('td', recursive=False)
            if tds and len(tds) > 1:
                if tds[1].get_text().find(self.unknown) == -1:
                    self.get_or_create_station(line, tds[0])



# -*- coding: utf-8 -*-
import re
import requests
from bs4 import BeautifulSoup

from metro.parser.base import BaseEnDataProvider


class DataProvider(BaseEnDataProvider):
    metro_lines_src = u"%s/wiki/Tokyo_subway" % BaseEnDataProvider.base_en_url
    lines_buf = []

    def download_lines(self):
        html = self.create_dom(self.metro_lines_src)
        table = html.find_all('table')[2]
        for row in table.find_all('tr'):
            tds = row.find_all('td', recursive=False)
            if tds and len(tds) == 5:
                self.lines_buf.append(dict(
                    url=tds[3].find('a').attrs['href'],
                    color=self.prep_color(tds[0]),
                    title=self.prep_title(tds[3]),
                    number=self.get_el_text(tds[2])\
                                .replace(u'Line ', '')\
                                .strip()
                ))

    def download_stations(self):
        # very slow (13 http queries, i preffer scrapy, but...)
        for line in self.lines_buf:
            # create line
            line_inst = self.get_or_create_line(
                line['number'], line['title'], line['color']
            )
            print u'Download stations for line number %s' % line['number']
            # download stations
            url = BaseEnDataProvider.base_en_url + line['url']
            html = self.create_dom(url)
            table = html.find('table', class_='wikitable')
            for row in table.find_all('tr', recursive=False):
                tds = row.find_all('td', recursive=False)
                th = row.find('th')
                if tds and len(tds) > 1:
                    index = 0 if th else 1
                    self.get_or_create_station(
                        line_inst, self.get_el_text(tds[index])
                    )



# -*- coding: utf-8 -*-
import requests
from django.db import models
from django.utils.translation import ugettext as _


class MetroLine(models.Model):
    title = models.CharField(
        max_length=255,
        verbose_name=_(u'Title')
    )
    color = models.CharField(
        max_length=255,
        verbose_name=_(u'Color'),
        blank=True,
        null=True,
    )
    number = models.IntegerField(
        verbose_name=_(u'Line number'),
        null=True,
        blank=True
    )
    icon = models.ImageField(
        upload_to='metro/',
        verbose_name=_(u'Icon'),
        blank=True,
        null=True,
    )

    @classmethod
    def get_all(cls):
        items = cls.objects.all()
        result = dict()
        for item in items:
            result[ item.number ] = item
        return result

    def get_admin_color(self):
        return u'<div style="width: 60px; height: 20px; '\
                'background: %s"></div>' % self.color
    get_admin_color.allow_tags = True
    get_admin_color.short_description = _(u'Color')

    def __unicode__(self):
        return unicode(self.title)

    class Meta:
        ordering = ['number']
        verbose_name = _(u'Metro line')
        verbose_name_plural = _(u'Metro lines')


class Metro(models.Model):
    line = models.ForeignKey(
        MetroLine,
        verbose_name=_(u'Metro line')
    )
    title = models.CharField(
        max_length=255,
        verbose_name=_(u'Title')
    )
    order = models.PositiveIntegerField(
        verbose_name=_(u'Order'),
        default=0,
    )

    def __unicode__(self):
        return unicode(self.title)

    class Meta:
        ordering = ['order']
        verbose_name = _(u'Metro station')
        verbose_name_plural = _(u'Metro stations')



# -*- coding: utf-8 -*-
import requests
from bs4 import BeautifulSoup

from metro.parser.base import BaseRuDataProvider


class DataProvider(BaseRuDataProvider):
    metro_lines_src = u"http://ru.wikipedia.org/wiki/Модуль:MoscowMetro#ColorByNum"
    metro_stations_src = u"http://ru.wikipedia.org/w/index.php?title=\
                           Список_станций_Московского_метрополитена"

    def download_lines(self):
        html = self.create_dom(self.metro_lines_src)
        table = html.find('table')
        for i, row in enumerate(table.find_all('tr')):
            if i == 0:
                continue
            number = 0
            for j, cell in enumerate(row.find_all('td')):
                value = cell.string
                if j == 0:
                    if value and value.isdigit():
                        number = int(value)
                elif j == 1:
                    title = value
                elif j == 2:
                    color = value
            if number > 0:
                self.get_or_create_line(number, title, '#' + color)

    def download_stations(self):
        html = self.create_dom(self.metro_stations_src)
        table = html.find('table', 'wikitable')
        lines = self.line_model.get_all()
        for i, row in enumerate(table.find_all('tr')):
            if i == 0:
                continue
            for j, cell in enumerate(row.find_all('td')):
                if j == 0:
                    line = 0
                    value = cell.find('span', 'sortkey').string
                    if value and value.isdigit():
                        line = int(value)
                elif j == 1:
                    title = cell.find('span').string
            try:
                line_inst = lines[line]
            except KeyError:
                continue
            self.get_or_create_station(line_inst, title)



# -*- coding: utf-8 -*-
import re
import requests
from bs4 import BeautifulSoup

from metro.parser.base import BaseRuDataProvider


class DataProvider(BaseRuDataProvider):
    metro_data_src = u"http://ru.wikipedia.org/wiki/\
                       Список_станций_Нижегородского_метрополитена"
    td_count = 4

    def download_all(self):
        i = 0
        line_trs = []
        line_data = dict()
        html = BeautifulSoup(requests.get(self.metro_data_src).content)
        headers = html.find_all('h2')
        for head in headers:
            if self.line_word in head.get_text():
                i += 1
                table = head.findNext('table')
                td = table.find('td')
                line = self.get_or_create_line(i, td)
                all_trs = table.find_all('tr', recursive=False)
                for tr in all_trs:
                    tds = tr.find_all('td', recursive=False)
                    if len(tds) == self.td_count:
                        self.get_or_create_station(line, tds[0])



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Metro',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('title', models.CharField(max_length=255, verbose_name='\u041d\u0430\u0437\u0432\u0430\u043d\u0438\u0435')),
                ('order', models.PositiveIntegerField(default=0, verbose_name='\u0421\u043e\u0440\u0442\u0438\u0440\u043e\u0432\u043a\u0430')),
            ],
            options={
                'ordering': ['order'],
                'verbose_name': '\u0421\u0442\u0430\u043d\u0446\u0438\u044f \u043c\u0435\u0442\u0440\u043e',
                'verbose_name_plural': '\u0421\u0442\u0430\u043d\u0446\u0438\u0438 \u043c\u0435\u0442\u0440\u043e',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='MetroLine',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('title', models.CharField(max_length=255, verbose_name='\u041d\u0430\u0437\u0432\u0430\u043d\u0438\u0435')),
                ('color', models.CharField(max_length=255, null=True, verbose_name='\u0426\u0432\u0435\u0442', blank=True)),
                ('number', models.IntegerField(help_text='\u0420\u0435\u0430\u043b\u044c\u043d\u044b\u0439 \u043d\u043e\u043c\u0435\u0440 \u043b\u0438\u043d\u0438\u0438', null=True, verbose_name='\u041d\u043e\u043c\u0435\u0440 \u043b\u0438\u043d\u0438\u0438', blank=True)),
                ('icon', models.ImageField(upload_to=b'metro/', null=True, verbose_name='\u0418\u043a\u043e\u043d\u043a\u0430', blank=True)),
            ],
            options={
                'ordering': ['number'],
                'verbose_name': '\u041b\u0438\u043d\u0438\u044f \u043c\u0435\u0442\u0440\u043e',
                'verbose_name_plural': '\u041b\u0438\u043d\u0438\u0438 \u043c\u0435\u0442\u0440\u043e',
            },
            bases=(models.Model,),
        ),
        migrations.AddField(
            model_name='metro',
            name='line',
            field=models.ForeignKey(verbose_name='\u041b\u0438\u043d\u0438\u044f \u043c\u0435\u0442\u0440\u043e', to='metro.MetroLine'),
            preserve_default=True,
        ),
    ]



# -*- coding: utf-8 -*-
import re
import requests
from bs4 import BeautifulSoup

from metro.parser.base import BaseEnDataProvider


class DataProvider(BaseEnDataProvider):
    metro_lines_src = u"%s/wiki/London_Underground" % BaseEnDataProvider.base_en_url
    lines_buf = []

    def download_lines(self):
        html = self.create_dom(self.metro_lines_src)
        table = html.find('table', class_='wikitable')
        i = 0
        for row in table.find_all('tr', recursive=False):
            tds = row.find_all('td', recursive=False)
            if tds and len(tds) > 1:
                i += 1
                self.lines_buf.append(dict(
                    url=tds[0].find('a').attrs['href'],
                    # very bad
                    color=self.prep_color(tds[1])\
                        .replace('color:', '')\
                        .replace('white', '')\
                        .replace('black', '')\
                        .strip(),
                    title=self.prep_title(tds[0]),
                    number=i
                ))

    def download_stations(self):
        for line in self.lines_buf:
            # create line
            line_inst = self.get_or_create_line(
                line['number'], line['title'], line['color']
            )
            print u'Download stations for line number %s' % line['number']
            # download stations
            url = BaseEnDataProvider.base_en_url + line['url']
            html = self.create_dom(url)
            tables = html.find_all('table', class_='wikitable')
            for table in tables:
                for row in table.find_all('tr', recursive=False):
                    tds = row.find_all('td', recursive=False)
                    if tds and len(tds) > 1:
                        self.get_or_create_station(
                            line_inst, self.get_el_text(tds[0])
                        )



# -*- coding: utf-8 -*-
from django.core.management.base import BaseCommand, CommandError

from metro.models import Metro, MetroLine


class Command(BaseCommand):
    def handle(self, *args, **options):
        MetroLine.objects.all().delete()
        Metro.objects.all().delete()



# -*- coding: utf-8 -*-
import re
import requests
from bs4 import BeautifulSoup

from metro.parser.base import BaseRuDataProvider


class DataProvider(BaseRuDataProvider):
    metro_data_src = u"http://ru.wikipedia.org/w/index.php?title=\
                       Список_станций_Минского_метрополитена"

    def download_all(self):
        self.parse_usual_big_table(station_td_count=5)



# -*- coding: utf-8 -*-
from django.conf import settings
from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class MetroConfig(AppConfig):
    name = u'metro'
    verbose_name = getattr(settings, 'METRO_APP_TITLE', _(u'Metro'))



# -*- coding: utf-8 -*-
from importlib import import_module
from django.conf import settings

from metro.models import Metro, MetroLine

provider = None

# import data provider
try:
    module = import_module(
        'metro.parser.providers.%s' % (
            getattr(settings, 'METRO_PROVIDER', 'moscow').lower()
        )
    )
    provider = module.DataProvider(
        station_model=Metro, line_model=MetroLine
    )
except ImportError:
    raise NotImplementedError(
        u'Provider for this city does not implemented'
    )



# -*- coding: utf-8 -*-
from django.core.management.base import BaseCommand, CommandError

from metro.parser import provider


class Command(BaseCommand):
    def handle(self, *args, **options):
        provider.download_all()
        print 'Done'



# -*- coding: utf-8 -*-
import re
import requests
from bs4 import BeautifulSoup

from metro.parser.base import BaseRuDataProvider


class DataProvider(BaseRuDataProvider):
    metro_data_src = u"http://ru.wikipedia.org/wiki/\
                       Линии_и_станции_Петербургского_метрополитена"
    header_marker = u'линия'
    skip_from = 6 # 6, 7, 8 lines is under construction
    title_re = re.compile(ur'.*?(\d{1,2}).*?\((.*?)\)', re.U | re.S)

    def download_all(self):
        html = self.create_dom(self.metro_data_src)
        headers = html.find_all(class_='mw-headline')
        lines = self.line_model.get_all()
        for row in headers:
            header = row.find('font')
            if header and header.string.lower().find(self.header_marker) > -1:
                matches = self.title_re.match(header.string)
                if matches:
                    # extract all line data
                    line_color = header['color']
                    line_number = int(matches.group(1))
                    if line_number >= self.skip_from:
                        continue
                    line_title = matches.group(2)
                    if not line_number in lines:
                        line = self.line_model.objects.create(
                            number=line_number,
                            color=line_color,
                            title=line_title
                        )
                    else:
                        line = lines[line_number]
                    # extract stations
                    stations_ul = header.parent.parent.find_next_sibling('ul')
                    if stations_ul:
                        for item in stations_ul.find_all('li'):
                            self.get_or_create_station(line, item.string)



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages
from metro import __version__


readme = 'README.md'
try:
    import pypandoc
    read = lambda: pypandoc.convert(readme, 'rst')
except ImportError:
    import os
    read = lambda: os.path.join(
        os.path.dirname(os.path.abspath(__file__)), readme
    )


setup(
    name='django-metro',
    version=__version__,
    description="""Basic metro models for Django 1.7+, \
    plus the parser that fills models with actual data \
    (lines numbers, lines titles, lines colors, the names \
    of the lines, the station names) from various data \
    sources (primary - Wikipedia).
    """,
    long_description=read(),
    url='https://github.com/xfenix/django-russian-metro',
    author='Xfenix',
    author_email='ad@xfenix.ru',
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ],
    zip_safe=True,
    install_requires=[
        'Django>=1.7',
        'BeautifulSoup4',
        'requests',
    ],
)



# -*- coding: utf-8 -*-
from django.contrib import admin

from .models import MetroLine, Metro


# try to use sortable stacked inline from django-suit
try:
    from suit.admin import SortableStackedInline
    class StackedAdmin(SortableStackedInline): pass
except ImportError:
    class StackedAdmin(admin.StackedInline): pass


class MetroAdminInline(StackedAdmin):
    model = Metro
    extra = 0
    list_display = ('title', 'line', )
    list_display_links = ('title', )
    list_filter = ('title', 'line')
    search_fields = ('title', )


class MetroLineAdmin(admin.ModelAdmin):
    inlines = [MetroAdminInline, ]
    list_display = ('number', 'title', 'get_admin_color')
    list_display_links = ('number', 'title', 'get_admin_color')
    list_filter = ('number', 'title')
    search_fields = ('number', 'title')


admin.site.register(MetroLine, MetroLineAdmin)



# -*- coding: utf-8 -*-
from metro.parser.base import BaseRuDataProvider


class DataProvider(BaseRuDataProvider):
    def download_all(self):
        # yeap, Omsk
        self.get_or_create_station(
            self.get_or_create_line(
                1, u'Первая', 'red'
            ),
            u'Библиотека имени Пушкина'
        )
